// ezrpm -- By Warren E. Downs
// August 23, 1999
// License: GPL (See www.fsf.org)

// Purpose: ezrpm takes one or more .rpm filenames as parameters, and installs,
// them handling all dependencies by installing *additional* .rpms as necessary
// (Additional .rpms are located in the same directories as used by specified rpms,
// or in one of the directories in the environment variable, RPMPATH).
// Process: This is done by querying all .rpm files in all determined source directories,
// to determine the file/packages/capability dependency tree.  Only the minimum
// packages necessary to install the specified files, are installed.

// Handling of removable media:
// 	  If paths are prefixed with [Volume Name], the user is prompted to insert
//    "Volume Name" before that path is checked.  This prompting is graphical
//    if the user appears to be running X, text-based otherwise, unless
//	  overridden with --nogui or --gui, accordingly.

// Needed for snprintf()
#define _GNU_SOURCE

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>

#include "ezrpmcls.h"
#include "comprpm.h"

#define  min(a,b)	(((a)<(b))?(a):(b))
#define  max(a,b)	(((a)>(b))?(a):(b))

// Default options
bool				gui=false;
bool				opinstall=true;
bool				force=false;
bool				opstdin=false;

// Other options
Tdir				*srcdirs=NULL;
Tdir				*srcdirs_tail=NULL;
Trpm				*rpms=NULL;
Trpm				*rpms_tail=NULL;

// Function prototypes
void help(char *name);
void addrpmdir(char *rpmpath);
bool addrpm(char *rpmfile);
char *pathof(char *pathname);
void deps(char *rpmfile);
void locate_rpm(char *rpmfile, char *fullrpm, int size);
char *find_file(char *filename);
char *find_pkgcap(char *pkgcap);
bool ispkg(char *pkgname, char *pkgcap);
Tcapability *caplist(char *rpmpath);
Tfile *filelist(char *rpmpath);
Tpackage *dirlist(char *dirname);
void install(char *rpm);

// Functions
void main(int argc, char **argv) {
	if(argc==1) {
		help(argv[0]);
		}
	else {
		int cc;
        int digit_optind = 0;

        // Set default to *terminate* rather than *resume* interrupted system calls.
        // This is necessary so fgets() will return when reading from a pipe and
        // receiving SIGCHLD.  (Otherwise the program gets stuck in "uninterruptible sleep").
		if(siginterrupt(SIGCHLD,1)==-1) {
			fprintf(stderr,"Signal %d is invalid.\n",SIGCHLD);
			}
		if(siginterrupt(SIGSEGV,1)==-1) {
			fprintf(stderr,"Signal %d is invalid.\n",SIGSEGV);
			}
        while (1) {
        	int this_option_optind = optind ? optind : 1;
            int option_index=0;
            static struct option long_options[]={
             	{"help",  0, 0, 0},
              	{"gui",   0, 0, 0},
              	{"nogui", 0, 0, 0},
              	{"noinstall", 0, 0, 0},
              	{"force", 0, 0, 0},
              	{"stdin", 0, 0, 0},
              	{0, 0, 0, 0}
             	};
            cc=getopt_long(argc,argv,"hgnf",long_options,&option_index);
            if(cc == -1) { break; }
            else if(cc=='h') { help(argv[0]); 	}
            else if(cc=='g') { gui=true ;     	}
            else if(cc=='n') { opinstall=false; }
            else if(cc=='f') { force=true;		}
            else if(cc==0) {
            	if(strcmp(long_options[option_index].name,"help")==0) {
					help(argv[0]);
            		}
            	else if(strcmp(long_options[option_index].name,"nogui")==0) {
            		gui=false;
            		}
            	else if(strcmp(long_options[option_index].name,"gui")==0) {
            		gui=true;
            		}
            	else if(strcmp(long_options[option_index].name,"noinstall")==0) {
            		opinstall=false;
            		}
            	else if(strcmp(long_options[option_index].name,"force")==0) {
            		force=true;
            		}
            	else if(strcmp(long_options[option_index].name,"stdin")==0) {
            		opstdin=true;
            		}
                if(optarg) { printf(" with arg %s", optarg); }
                }
            else if(cc=='?') {
            	exit(1);
            	}
            else {
               	printf ("?? getopt returned character code 0%o ??\n",cc);
               	exit(2);
              	}
          	}
	
	    if((optind==argc) && (!opstdin)) {
	    	fprintf(stderr,"No packages supplied.\n");
	    	exit(3);
	    	}
	
	    // FIXME: Eventually, opstdin option should trigger reading packages from
	    //        standard input instead of from the argument list.
		char *rpmpath;
		for(int ii=optind; ii<argc; ii++) {
			rpmpath=pathof(argv[ii]);
			addrpmdir(rpmpath);	// Adds path of given file to .rpm path list
			}
		rpmpath=getenv("RPMPATH");
		for(int ii=0; rpmpath && *rpmpath; ii++) {
			char *qq=strchr(rpmpath,':');
			if(qq) { *qq=0; qq++; }
			addrpmdir(rpmpath);	// Adds path to .rpm path list
			if(qq) { rpmpath=qq; }
			else   { rpmpath=""; }
			}
		printf("RPM paths:\n");
		for(Tdir *dir=srcdirs; dir; dir=dir->next) {
			printf("  %-57s %-20s\n",dir->name,dir->prompt);
			}
		printf("Checking dependencies...\n");
		for(int ii=optind; ii<argc; ii++) {
			///**/printf("  %s:\n",argv[ii]);
			deps(argv[ii]);
			}
		// Reverse package order (so deepest dependencies get installed first)
		Trpm *pp=NULL;
		Trpm *qq=rpms->next;
		rpms->next=NULL;
		for(Trpm *rpm=rpms; qq; rpm=pp) {
			pp=qq; qq=pp->next; pp->next=rpm;
			}
		if(opinstall) {
			printf("\nInstalling packages...\n");
			}
		else {
			printf("\nPackage dependencies:\n");
			}
		for(Trpm *rpm=rpms_tail; rpm; rpm=rpm->next) {
			printf("  %s\n",rpm->name);
			if(opinstall) {
				install(rpm->name);
				}
			}
		}
	}

// FIXME: We could have an option to only install/display packages that weren't
//        already installed.
void help(char *name) {
	fprintf(stderr,"Usage:\n");
	fprintf(stderr,"  %s [filenames] {--gui | --nogui} {--noinstall} {--force} | --help \n",name);
	fprintf(stderr,"    (Installs .rpms, resolving dependencies from install dirs or $RPMPATH.\n");
	fprintf(stderr,"    If an RPMPATH item is prefixed with [VolName], prompt user for volume\n");
	fprintf(stderr,"    VolName).\n");
	fprintf(stderr,"    Legend:\n");
	fprintf(stderr,"      . = dependency check, + = file list, * = capability list\n");
	exit(4);
	}
	
// Returns the path portion of the filename
char *pathof(char *pathname) {
	static char path[1024];
	strncpy(path,pathname,sizeof(path)); path[sizeof(path)-1]=0;
	char *qq=strrchr(path,'/');
	if(qq) { *qq=0;				}
	else   { strcpy(path,".");	}
	return path;
	}

// Add given path to RPM path list if not already in it.
void addrpmdir(char *rpmpath) {
	char			prompt[256];
	if(!rpmpath) return;
	*prompt=0;
	if(*rpmpath=='[') {
		char *qq=strchr(rpmpath,']');
		if(qq) {
			*qq=0; strcpy(prompt,rpmpath+1); rpmpath=qq+1;
			}
		}
	for(Tdir *dir=srcdirs; dir; dir=dir->next) {
		if(strcmp(dir->name,rpmpath)==0) {
			if(!dir->prompt[0] && prompt[0]) { strcpy(dir->prompt,prompt); }
			return;							// Already in the list
			}
		}
	Tdir *dir=new Tdir(rpmpath,prompt);
	if(srcdirs) { srcdirs_tail=srcdirs_tail->add(dir); }
	else		{ srcdirs_tail=srcdirs=dir;			   }
	}

// Add given file RPM file list (to be installed) if not already in it.
// Returns true if RPM was already in list; false otherwise.
bool addrpm(char *rpmfile) {
	if(!rpmfile) return false;
	for(Trpm *rpm=rpms; rpm; rpm=rpm->next) {
		if(strcmp(rpm->name,rpmfile)==0) {
			return true;						// Already in the list
			}
		}
	Trpm *rpm=new Trpm(rpmfile);
	if(rpms) { rpms_tail=rpms_tail->add(rpm); }
	else	 { rpms_tail=rpms=rpm;			  }
	/**/printf("\n%s\n",rpmfile);
	return false;
	}

// Build dependencies for this rpm
// The specified rpm file is added to the list of RPMs to be installed.
// The data returned by rpm is in the following form:
// (Example from MAKEDEV-2.5-2mdk.noarch.rpm)
/*
    /bin/sh						// File
    fileutils					// Package (if it exists) or capability (seach packages)
    shadow-utils >= 970616-7	// Package version requirement
    shadow-utils				// Package/capability
    /bin/sh						// File
    /bin/bash					// File
*/
// Determining whether a requirement is for a package or a capability:
// We first extract the name up to the first space (if any).
//
// If a specific version is required (e.g. "pkg >= 970616-7", or
//   "pkg = 970616-7", etc.), we look for a matching package (capabilities don't
//   have versions).  We perform this search in all specified search dirs.
//   In order to match, the package must have the name as a prefix, followed by
//   a dash ("-"), and an equivalent (or greater, lesser, etc.) version.
//
// Else,
//   For each search directory provided,
//     If a package exists with the name in question as a prefix, immediately
//       followed by a dash ("-") and a digit (0-9), it is treated as the
//       desired package.
//     Else, we search all packages in the current directory we're processing,
//       using "rpm -qp --provides" on each package, to determine if any provides
//       the required capability.
void deps(char *rpmfile) {
	FILE 			*fp;
	char			buf[256];
	char			rpm[256];
	
	printf("."); fflush(stdout);
	locate_rpm(rpmfile,rpm,sizeof(rpm));
	if(!rpm[0]) {
		fprintf(stderr,"Package %s not found in RPMPATH or listed dirs!\n",rpmfile);
		return;
		}
		
	if(addrpm(rpm)) { return; }	// This RPM was already in list; don't check deps again
	
	snprintf(buf,sizeof(buf),"rpm -qpR %s",rpm);
	///**/printf("****%s\n",buf);
	if((fp=popen(buf,"r"))==NULL) {
		fprintf(stderr,"Error in deps():popen %s\n",strerror(errno)); return;
		}
	while(!feof(fp)) {
		if(fgets(buf,sizeof(buf),fp)) {
			char *pkg=NULL;
			char *pp=NULL;
			if(buf[0]) { pp=&buf[strlen(buf)-1]; }
			while(pp && isspace(*pp)) { *pp=0; pp--; }
			///**/printf("    %s\n",buf);
			if(pkg=find_file(buf)) {			// Dependency is a file
				deps(pkg);						// Add this package to list of RPMs to install
				}
			else if(pkg=find_pkgcap(buf)) {		// Dependency is a package or capability
				deps(pkg);						// Add this package to list of RPMs to install
				}
			else {								// Dependency was not found
				fprintf(stderr,"\nMissing dependency: %s\n",buf);				
				}
			}
		}
	if(pclose(fp)==-1) {
		fprintf(stderr,"Error in deps():pclose %s\n",strerror(errno));
		}
	}
	
void locate_rpm(char *rpmfile, char *fullrpm, int size) {
	if((rpmfile[0]=='.') || (rpmfile[0]=='/')) {	// If this is a fully-specified path...
		strncpy(fullrpm,rpmfile,size); return;
		}
   	for(Tdir *dir=srcdirs; dir; dir=dir->next) {
   		if(!dir->pkgs) {						// Get dirlist for this directory
   			///**/printf("Scanning %s...",dir->name);
   			dir->pkgs=dirlist(dir->name);
   			}
		///**/printf("Scanning for packages in %s\n",dir->name);
   		for(Tpackage *pkg=dir->pkgs; pkg; pkg=pkg->next) {
   			if(strcmp(pkg->name,rpmfile)==0) {
   				snprintf(fullrpm,size,"%s/%s",dir->name,pkg->name);
   				///**/printf("Found %s\n",fullrpm);
   				return;
   				}
   			}
		}
	if(size>0) { fullrpm[0]=0; }
	return;
	}
	
// Search the dependency tree to locate a package containing this file.
// If a given directory's 'pkgs' is NULL, get the directory list.
// If a given package's 'files' is NULL, get the package's file list.
char *find_file(char *filename) {
	if(!filename || (*filename!='/')) { return NULL; }
   	for(Tdir *dir=srcdirs; dir; dir=dir->next) {
   		if(!dir->pkgs) {						// Get dirlist for this directory
   			///**/printf("Scanning %s...\n",dir->name);
   			dir->pkgs=dirlist(dir->name);
   			}
		///**/printf("Scanning package contents in %s\n",dir->name);
   		for(Tpackage *pkg=dir->pkgs; pkg; pkg=pkg->next) {
   			if(!pkg->files) {				// Get files list for this package
   				char path[256];
   				snprintf(path,sizeof(path),"%s/%s",dir->name,pkg->name);
   				pkg->files=filelist(path);
   				}
   			for(Tfile *file=pkg->files; file; file=file->next) {
   				if(strcmp(file->name,filename)==0) {
   					return pkg->name;
   					}
   				}
   			}
   		}
	return NULL;
	}
	
// Search the dependency tree to locate a package by this name, or
// providing this capability.
// If a given directory's 'pkgs' is NULL, get the directory list.
// If a given package's 'caps' is NULL, get the package's capability list.
char *find_pkgcap(char *pkgcap) {
	if(!pkgcap) { return NULL; }
   	for(Tdir *dir=srcdirs; dir; dir=dir->next) {
   		if(!dir->pkgs) {						// Get dirlist for this directory
   			///**/printf("Scanning %s...\n",dir->name);
   			dir->pkgs=dirlist(dir->name);
   			}
		///**/printf("Scanning packages and capabilities in %s\n",dir->name);
   		for(Tpackage *pkg=dir->pkgs; pkg; pkg=pkg->next) {
   			if(ispkg(pkg->name,pkgcap)) {
   				return pkg->name;
   				}
   			if(!pkg->caps) {					// Get capability list for this package
   				char path[256];
   				snprintf(path,sizeof(path),"%s/%s",dir->name,pkg->name);
   				pkg->caps=caplist(path);
  				}
   			for(Tcapability *cap=pkg->caps; cap; cap=cap->next) {
   				if(strcmp(cap->name,pkgcap)==0) {
   					return pkg->name;
   					}
   				}
   			}
   		}
	return NULL;
	}
	
bool ispkg(char *pkgname, char *pkgcap) {
	char			buf1[256];
	char			ver1[256];
	char			pkgreq[256];
	int				rel=2;		// Default to ">="
	char			*pp=NULL;
	char			*qq=NULL;
	char			*ver="";
	
	strncpy(pkgreq,pkgcap,sizeof(pkgreq)); pkgreq[sizeof(pkgreq)-1]=0;
	// Is this a comparative versioned requirement? (<, >, =, <=, >=)
	if(pp=strstr(pkgreq,"<="))      { rel=1; ver=pp+2; }
	else if(pp=strstr(pkgreq,"=<")) { rel=1; ver=pp+2; }
	else if(pp=strstr(pkgreq,">=")) { rel=2; ver=pp+2; }
	else if(pp=strstr(pkgreq,"=>")) { rel=2; ver=pp+2; }
	else if(pp=strstr(pkgreq,"==")) { rel=3; ver=pp+2; }
	else if(pp=strstr(pkgreq,"="))  { rel=3; ver=pp+1; }
	else if(pp=strstr(pkgreq,"<"))  { rel=4; ver=pp+1; }
	else if(pp=strstr(pkgreq,">"))  { rel=5; ver=pp+1; }
	// If yes, extract the required version
	if(pp) {
		qq=strchr(pkgreq,' ');
		if(qq && (qq<pp)) { *qq=0; }
		while(isspace(*ver)) { ver++; }
		}
	base_name(pkgname,buf1,sizeof(buf1));
	version(pkgname,ver1,sizeof(ver1));
	if(strcmp(buf1,pkgreq)==0) {	// We have matching packages.  Do the versions match?
		char		fakepkg[256];
		char		ver2[256];
		snprintf(fakepkg,sizeof(fakepkg),"-%s.rpm",ver);
		version(fakepkg,ver2,sizeof(ver2));
		//printf("%s (%s) --> %s (%s '%s')?\n",buf1,ver1,pkgreq,relstr,ver2);
    	switch(rel) {
    		case 1: return strcmp(ver1,ver2)<=0;
    		case 2: return strcmp(ver1,ver2)>=0;
    		case 3: return strcmp(ver1,ver2)==0;
    		case 4: return strcmp(ver1,ver2)<0;
    		case 5: return strcmp(ver1,ver2)>0;
    		}
		
		return true;				// Say yes for now...
		}
	return false;
	}

// Returns a new list of capabilities provided by the given RPM file
Tcapability *caplist(char *rpmpath) {
	FILE 			*fp;
	char			buf[256];
	Tcapability		*caps=NULL;
	Tcapability		*caps_tail=NULL;
	
	printf("+"); fflush(stdout);
	snprintf(buf,sizeof(buf),"rpm -qp --provides %s",rpmpath);
	///**/printf("****%s\n",buf);
	if((fp=popen(buf,"r"))==NULL) {
		fprintf(stderr,"Error in caplist():popen %s\n",strerror(errno)); return NULL;
		}
//	int cc;
//	if((cc=fgetc(fp))!=EOF) {
//		ungetc(cc,fp);
       	while(!feof(fp)) {
        	if(fgets(buf,sizeof(buf),fp)) {
        		char *pkg=NULL;
				char *pp=NULL;
				if(buf[0]) { pp=&buf[strlen(buf)-1]; }
				while(pp && isspace(*pp)) { *pp=0; pp--; }
        		///**/printf("    %s\n",buf);
        		Tcapability *cap=new Tcapability(buf);
        		if(caps) { caps_tail=caps_tail->add(cap); }
        		else	 { caps_tail=caps=cap;			  }
        		}
        	}
		if(pclose(fp)==-1) {
			fprintf(stderr,"Error in caplist():pclose %s\n",strerror(errno));
			}
//        }
//	else {
//		/**/printf("No capabilities.  Not calling pclose()\n");
//		pclose(fp);
//		}
	return caps;
	}
	
// Returns a new list of files contained in the given RPM file
Tfile *filelist(char *rpmpath) {
	FILE 			*fp;
	char			buf[256];
	Tfile			*files=NULL;
	Tfile			*files_tail=NULL;
	
	printf("*"); fflush(stdout);
	snprintf(buf,sizeof(buf),"rpm -qpl %s",rpmpath);
	///**/printf("****%s\n",buf);
	if((fp=popen(buf,"r"))==NULL) {
		fprintf(stderr,"Error in filelist():popen %s\n",strerror(errno)); return NULL;
		}
	while(!feof(fp)) {
		if(fgets(buf,sizeof(buf),fp)) {
			char *pkg=NULL;
			char *pp=NULL;
			if(buf[0]) { pp=&buf[strlen(buf)-1]; }
			while(pp && isspace(*pp)) { *pp=0; pp--; }
			///**/printf("    %s\n",buf);
			Tfile *file=new Tfile(buf);
			if(files) { files_tail=files_tail->add(file); }
			else	  { files_tail=files=file;			  }
			}
		}
	if(pclose(fp)==-1) {
		fprintf(stderr,"Error in filelist():pclose %s\n",strerror(errno));
		}
	return files;
	}
	
// Returns a new list of packages contained in the given directory
Tpackage *dirlist(char *dirname) {
	DIR 			*dir;	
	Tpackage		*pkgs=NULL;
	Tpackage		*pkgs_tail=NULL;
	struct dirent	*diritem;
	if((dir=opendir(dirname))==NULL) {
		fprintf(stderr,"opendir(%s): %s\n",dirname,strerror(errno)); return NULL;
		}
	while(diritem=readdir(dir)) {
		int			len=strlen(diritem->d_name);
		char 		*rpmptr=&diritem->d_name[len-4];
		if((len>4) && (strcasecmp(rpmptr,".rpm")==0)) {
			Tpackage *pkg=new Tpackage(NULL,diritem->d_name);
			if(pkgs) { pkgs_tail=pkgs_tail->add(pkg); }
			else	 { pkgs_tail=pkgs=pkg;			  }
			}
		}
	if(closedir(dir)==-1) {
		fprintf(stderr,"closedir(): %s\n",strerror(errno));
		}
	return pkgs;
	}

// Install this rpm (all dependencies must already be resolved at this point)
void install(char *rpm) {
	char buf[1024];
	sprintf(buf,"rpm -Uvh %s %s",force?"--force":"",rpm);
	system(buf);
	}
